/* 
   - The contents of this file are subject to the Mozilla Public
   - License Version 1.1 (the "License"); you may not use this file
   - except in compliance with the License. You may obtain a copy of
   - the License at http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS
   - IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   - implied. See the License for the specific language governing
   - rights and limitations under the License.
   -
   - The Original Code is this file as it was released on
   - August 30, 2003 (rewrite from old source files).
   -
   - The Initial Developer of the Original Code is Henk-Johan van Rantwijk.
   - Portions created by Henk-Johan van Rantwijk are Copyright (C) 2003-2004
   - Henk-Johan van Rantwijk.  All Rights Reserved.
   -
*/

var gDebug = false; // change this to true to enable debug info on the console

var mzRDFStore = null; // global pointer to multizilla.rdf

if (!RDF) {
  var RDF = Components.classes[ "@mozilla.org/rdf/rdf-service;1" ]
                      .getService( Components.interfaces.nsIRDFService );
}
var mzRDFCUtils = Components.classes["@mozilla.org/rdf/container-utils;1"]
                            .getService(Components.interfaces.nsIRDFContainerUtils);

var mzTabIndex = 0; // global index to number of tabs that have been restored
var gSessionName = "";

/**
 * File mode bits
 * #define PR_IRWXU 00700  read, write, execute/search by owner
 * #define PR_IRUSR 00400  read permission, owner
 * #define PR_IWUSR 00200  write permission, owner
 * #define PR_IXUSR 00100  execute/search permission, owner
 * #define PR_IRWXG 00070  read, write, execute/search by group
 * #define PR_IRGRP 00040  read permission, group
 * #define PR_IWGRP 00020  write permission, group
 * #define PR_IXGRP 00010  execute/search permission, group
 * #define PR_IRWXO 00007  read, write, execute/search by others
 * #define PR_IROTH 00004  read permission, others
 * #define PR_IWOTH 00002  write permission, others
 * #define PR_IXOTH 00001  execute/search permission, others
 * 
 * File open bits
 * #define PR_RDONLY       0x01
 * #define PR_WRONLY       0x02
 * #define PR_RDWR         0x04
 * #define PR_CREATE_FILE  0x08
 * #define PR_APPEND       0x10
 * #define PR_TRUNCATE     0x20
 * #define PR_SYNC         0x40
 * #define PR_EXCL         0x80
*/

//////////////////////////////////////////////////////////////////////////
// This function returns a nsiURI of the users profile directory or null 
// in case of failure. It is used in getNewOrCurrentDSByFilename()
//////////////////////////////////////////////////////////////////////////
function getProfileURI()
{
  var profile = Components.classes["@mozilla.org/profile/manager;1"].getService(Components.interfaces.nsIProfileInternal);
  var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
  var URI = ioService.newFileURI(profile.getProfileDir(profile.currentProfile));
  return URI || null;
}

function initializeRDFStorageFile() 
{
  mzRDFStore = getNewOrCurrentDSByFilename("multizilla.rdf", true, null, false);  // "quickprefs.rdf" for stand-alone versions of QPM
}
////////////////////////////////////////////////////////////////////////////////
// This initializes or constructs a new RDF (multizilla.rdf) It depends on two other 
// function, getProfileURI() and mzWriteToFile() so keep that in mind. It 
// returns the datasource or null on failure
////////////////////////////////////////////////////////////////////////////////
function getNewOrCurrentDSByFilename(aFileName, createWhenMissing, RDFData, hasPath)
{
  var fileName = hasPath ? aFileName : (getProfileURI().spec + aFileName);
  var ioService = Components.classes['@mozilla.org/network/io-service;1']
                            .getService(Components.interfaces.nsIIOService);
  var fileHandler = ioService.getProtocolHandler("file")
                             .QueryInterface(Components.interfaces.nsIFileProtocolHandler);
  var file = fileHandler.getFileFromURLSpec(fileName)
                        .QueryInterface(Components.interfaces.nsILocalFile);
  fileName = fileHandler.getURLSpecFromFile(file)

  if (createWhenMissing && !file.exists()) {
    RDFData = "<?xml version=\"1.0\"?>\n" +
              "<RDF:RDF xmlns:MZ=\"http://multizilla.mozdev.org/rdf#\"\n" +
              "         xmlns:RDF=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">\n" +
              "  <!-- Empty -->\n" +
              "</RDF:RDF>\n";
    mzWriteToFile(file, RDFData);
    // return RDF.GetDataSourceBlocking(fileName) || null;
  }
  // return RDF.GetDataSource(fileName) || null;
  var ds = RDF.GetDataSourceBlocking(fileName) || null;
  // alert(ds); // [xpconect wrapped nsIRDFDataSource]
  return ds;
}
////////////////////////////////////////////////////////////////////////////////
// This function returns a RDFContainer based on the given datasource 
// 'aDatasource', the given type 'aType' and the given resource 'aResource'. It 
// will creates a new container when you call this function with 'true' for 
// 'aInit'. It returns null on failure.
////////////////////////////////////////////////////////////////////////////////
function mzConstructRDFContainer(aDataSource, aType, aResource, aSkipInitFlag)
{
  var container = Components.classes["@mozilla.org/rdf/container;1"]
                            .createInstance(Components.interfaces.nsIRDFContainer);

  if (aDataSource && aType && aResource) {
    try {
      switch(aType) {
        default:
        case "seq": mzRDFCUtils.MakeSeq(aDataSource, RDF.GetResource(aResource));
          break;
        case "bag": mzRDFCUtils.MakeBag(aDataSource, RDF.GetResource(aResource));
          break;
        case "alt": mzRDFCUtils.MakeAlt(aDataSource, RDF.GetResource(aResource));
          break;
      }
      if (!aSkipInitFlag)
        container.Init(aDataSource, RDF.GetResource(aResource));
    } catch (ex) {
      mzDebug(ex);
      return null;
    }
  }
  return container;
}

function openSessionManager(aPrefFlag, skipCheck)
{
  var tsmIDs = [ "multizilla:tsm-prefwindow", "multizilla:smproperties", 
                 "multizilla:select-target-window", "multizilla:tsm:renametabsession", 
                 "multizilla:sessionmanager" ];

  for (i in tsmIDs) {
    var dialogWin = getMostRecentWindow(tsmIDs[i], false);

    if (dialogWin) {
      dialogWin.focus();
      break;
    }
    else {
      if (i == tsmIDs.length-1) {
        window.openDialog("chrome://multiviews/content/tabsession-manager/tabSessionManager.xul",  
                          "multizilla:sessionmanager", "chrome, dialog, resizable=yes", (aPrefFlag ? "preferences" : ""), skipCheck);
        break;
      }
    }
  }
}

function getNextSessionNumber(aSessionNumber, aWindowID)
{
  while (aSessionNumber < 15) {
    var resource = "urn:session-"+aSessionNumber+":"+aWindowID+":tabs:root";
    if (!checkResource(mzRDFStore, resource))
      return aSessionNumber;
    aSessionNumber++;
  }
  return 0; // Maximum number of sessions reached!
}

function getNextWindowNumber(aSessionID, aWindowNumber)
{
  while (aWindowNumber < 15) {
    var resource = "urn:"+aSessionID+":window-"+aWindowNumber+":tabs:root";
    if (!checkResource(mzRDFStore, resource))
      return aWindowNumber;
    aWindowNumber++;
  }
  return 0; // Maximum number of sessions reached!
}

function getNextSessionID(aSessionNumber, aWindowID)
{
  while (aSessionNumber < 15) {
    var resource = "urn:session-"+aSessionNumber+":"+aWindowID+":tabs:root";
    if (!checkResource(mzRDFStore, resource))
      return "session-"+aSessionNumber;
    aSessionNumber++;
  }
  return "session-0"; // Maximum number of sessions reached!
}

function getWindowResourceID(aWindow)
{
  var windowIndex = 0;
  var checkedWindows = 0;
  var browserWindows = getNumberOfWindowsByType();
  var windowManagerDS = Components.classes['@mozilla.org/rdf/datasource;1?name=window-mediator'].getService(nsIWindowDataSource);

  while (checkedWindows <= browserWindows) {
    var resourceWindowID = "window-"+windowIndex;
    var resourceWindow = windowManagerDS.getWindowForResource(resourceWindowID);
    if (resourceWindow) {
      if (resourceWindow == aWindow)
        return resourceWindowID;
      else if ('getBrowser' in resourceWindow)
        checkedWindows++;
    }
    windowIndex++;
  }
  return null;
}
////////////////////////////////////////////////////////////////////////////////
// This function returns an array with datasource ID's of all open browser 
// (navigator:browser) windows, or null when it fails to locate any of them, 
// but that should never happen (just to prevent a JS strict error).
////////////////////////////////////////////////////////////////////////////////
function getWindowResourceIDList()
{
  var windowArray = new Array();
  var windowIndex = 0;
  var checkedWindow = 0;
  var browserWindows = getNumberOfWindowsByType();
  var windowManagerDS = Components.classes['@mozilla.org/rdf/datasource;1?name=window-mediator'].getService(nsIWindowDataSource);

  while (checkedWindow <= browserWindows) {
    var resourceWindowID = "window-" + windowIndex;
    var resourceWindow = windowManagerDS.getWindowForResource(resourceWindowID);

    if (resourceWindow) {
      if ('getBrowser' in resourceWindow) {
        windowArray.push(resourceWindowID);
        checkedWindow++;
        if (checkedWindow == browserWindows)
          break;
      }
    }
    windowIndex++;
  }
  if (windowArray.length)
    return windowArray;
  return null;
}
////////////////////////////////////////////////////////////////////////////////
// This function returns a number (window-X) for the given window. This number 
// is based on the position of the datasource ID in the array of open windows. 
// It should never returns -1 but that's only to prevent a JS strict warning.
////////////////////////////////////////////////////////////////////////////////
function getWindowIndexNumber(aWindow)
{
  var thisWindowID = getWindowResourceID(aWindow);
  var allWindowIDs = getWindowResourceIDList();

  for (index in allWindowIDs) {
    if (allWindowIDs[index] == thisWindowID)
      return index;
  }
  return -1;
}
////////////////////////////////////////////////////////////////////////////////
// This function returns a datasource ID (window-x) for the given window. This 
// ID is based on the position in the array of all open browser windows. It 
// returns 'window-x', or null, just to prevent a JS strict warning.
////////////////////////////////////////////////////////////////////////////////
function getRelativeWindowResourceID(aWindow)
{
  var thisWindowID = getWindowResourceID(aWindow);
  var allWindowIDs = getWindowResourceIDList();

  for (index in allWindowIDs) {
    if (allWindowIDs[index] == thisWindowID)
      return "window-"+index;
  }
  return null;
}

function getSessionContainer(aDataSource)
{
  return mzConstructRDFContainer(aDataSource, "seq", "urn:sessions:root", false);
}
////////////////////////////////////////////////////////////////////////////////
// This function returns an array with all matching windows for the given 
// session. IE "session-x" where 'x' is the session number to lookup for.
////////////////////////////////////////////////////////////////////////////////
function getWindowResourcesForSession(aDataSource, aSession)
{
  var matchingResources = new Array();
  var sessionContainer = mzRDFCUtils.MakeSeq(aDataSource, RDF.GetResource("urn:sessions:root")).GetElements();
  
  while (sessionContainer.hasMoreElements()) {
    var resource = sessionContainer.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
    if (resource.Value.match(aSession))
      matchingResources.push(resource.Value);
  }
  return matchingResources.sort();
}

function openWindowAndReturnWindowID()
{
  window.open();
  var allWindowIDs = getWindowResourceIDList();
  return allWindowIDs[allWindowIDs.length-1];
}

function openWindowAndReturnBrowser()
{
  // This function fails when it is called from a modal window!
  window.open();
  var allWindowIDs = getWindowResourceIDList();
  var windowID = allWindowIDs[allWindowIDs.length-1];
  var windowManagerDS = Components.classes['@mozilla.org/rdf/datasource;1?name=window-mediator'].getService(nsIWindowDataSource);
  var browserWin = windowManagerDS.getWindowForResource(windowID);

  if (browserWin)
    return browserWin.getBrowser();
  return null;
}

function saveTabSession(aForcedSaveFlag, aCloseWSFlag)
{
  if (gBrowser.mChromeHidden) // Skip chromeless windows
    return;

  var saveSessionPref = readMyPref("multizilla.tabs.save-session.behavior", "int", 3);
  var shouldSave = (saveSessionPref > 0 || aForcedSaveFlag);
  var tabs = getBrowser().browsers.length;
  // Skip if we're not enabled, all tabs are blank
  if (!shouldSave || areAllTabsBlank() || tabs == 1 && readMyPref("multizilla.browsersession.skip.save-single-tab", "bool", false))
    return;

  var sessionID = 0;
  var windowID = getRelativeWindowResourceID(window);

  if (windowID == null) {
    alert("Error in saveTabSession()");
    return;
  }
  var resource = "urn:session-"+sessionID+":"+windowID+":tabs:root";
  var previouslyStored = checkResource(mzRDFStore, resource) ? true : false;
  var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
  // var checkboxMsg = mzGetTextFromBundle("saveTabSessionDontAskAgain");
  var checkbox = {value:0};
  var result = 0;
  var title;
  var text;
  // Fix for: http://mozdev.org/bugs/show_bug.cgi?id=4974
  var askConfirmation = (saveSessionPref == 3) ? true : false;
  var saveAsNewSession = (saveSessionPref == 2) ? true : false;

  /* if (askConfirmation) {
    return window.openDialog("chrome://multiviews/content/session-manager/tabSessionSaver.xul", 
                             "chrome, dependent, modal, dialog, centerscreen", "sessionSaver", 
                             window, saveAsNewSession);
    alert("gButtonPressed: " + gButtonPressed);
    return gButtonPressed;
  } */

  if (previouslyStored && askConfirmation) {
    title = mzGetTextFromBundle("previousTabSessionTitle");
    text = mzGetTextFromBundle("previousTabSessionText");
    const addButtonLabel = mzGetTextFromBundle("previousTabSessionAddButton");
    const overwriteButtonLabel = mzGetTextFromBundle("previousTabSessionOverwriteButton");
    result = promptService.confirmEx(window, title, text, 
                                    (promptService.BUTTON_POS_0 * promptService.BUTTON_TITLE_IS_STRING) +
                                    (promptService.BUTTON_POS_1 * promptService.BUTTON_TITLE_DONT_SAVE) +
                                    (promptService.BUTTON_POS_2 * promptService.BUTTON_TITLE_IS_STRING),
                                     addButtonLabel, null, overwriteButtonLabel, null, checkbox);
    if (result == 1) // 0 = new / 1 = don't save (ESC) / 2 = overwrite
      return;
    if (result == 0) {
      sessionID = getNextSessionID(sessionID, windowID);
      resource = "urn:"+sessionID+":"+windowID+":tabs:root";
    }
  }
  else if (!gBrowser.mRestoreBrowserSession && askConfirmation) {
    // This part will be skipped when we found a previously saved tab session with the same ID  
    // and when the browser was startup for an external link or external application.
    title = mzGetTextFromBundle("saveTabSessionTitle");
    text = mzGetTextFromBundle("saveTabSessionText");
    result = promptService.confirmEx(window, title, text, 
                                    (promptService.BUTTON_POS_0 * promptService.BUTTON_TITLE_SAVE) +
                                    (promptService.BUTTON_POS_1 * promptService.BUTTON_TITLE_DONT_SAVE),
                                     null, null, null, null, checkbox);
    if (result == 1) // 0 = save / 1 = don't save (ESC)
      return;
  }
  if (saveAsNewSession) { // Fix for: http://mozdev.org/bugs/show_bug.cgi?id=4974
    sessionID = getNextSessionID(sessionID, windowID);
    resource = "urn:"+sessionID+":"+windowID+":tabs:root";
  }
  storeGroupData(mzRDFStore, resource, null, aCloseWSFlag, true);
}

function storeGroupData(aDataSource, aResource, aBrowser, aCloseWSFlag, skipAboutBlank)
{
  if (skipAboutBlank == undefined)
    skipAboutBlank = true;

  removeAllResourcesFor(aDataSource, aResource);
  addTabDataSeqToDS(aDataSource, aResource, 0, aBrowser, skipAboutBlank);
  // Add resourceID for this tab session only when it is missing!
  if (!checkResourceForValue(aDataSource, "urn:sessions:root", aResource)) {
    var sessionContainer = getSessionContainer(aDataSource);
    sessionContainer.AppendElement(RDF.GetResource(aResource));
    mzFlushDataSource(aDataSource);
  }
  writeSessionName(aDataSource, aResource, aResource);
  mzFlushDataSource(aDataSource);

  if (!aCloseWSFlag)
    renameTabSession(aDataSource, aResource);
}

function addTabDataSeqToDS(aDataSource, aResource, startIndex, aBrowser, skipAboutBlank)
{
  if (!aResource)
    return -1;

  if (startIndex)
    aResource = aResource.replace(/:tabs:tab-\w*/, ":tabs:root");

  var tabContainer = mzConstructRDFContainer(aDataSource, "seq", aResource, false);

  if (!tabContainer)
    return -1;

  if (startIndex == 'undefined')
    startIndex = 0;

  aBrowser = aBrowser ? aBrowser : gBrowser;

  var tabList = aBrowser.mTabList;
  var tabs = aBrowser.mTabContainer.childNodes.length;
  var resourcePrefix = aResource.replace(/tabs:root/, "tabs");
  var mzRDF = "http://multizilla.mozdev.org/rdf#";
  var propertyArray = ["label", "custom", "firstpage", "image", "location", "interval", "selected", "ordinal", "allowPlugins", "allowJavascript", "allowMetaRedirects", "allowSubframes", "allowImages"];
  var defaultValues = ["about:blank", "false", "about:blank", "", "about:blank", "0", "false", "true", "true", "true", "true", "true"];

  for (var index = 0; index < tabs; index++) {
    var browser = aBrowser.getBrowserForTab(tabList[index]);
    var location = browser.webNavigation.currentURI.spec;

    if (skipAboutBlank && location == "about:blank")
      continue;

    var propertyValue = new Array();
    // Data collecting
    propertyValue.push(tabList[index].getAttribute("label"));
    propertyValue.push(tabList[index].hasAttribute("customLabel") ? "true" : "false");
    propertyValue.push(tabList[index].getAttribute("firstPage"));
    propertyValue.push(tabList[index].getAttribute("image"));
    propertyValue.push(location);
    propertyValue.push(tabList[index].mReloadInterval);
    propertyValue.push((tabList[index] == gBrowser.selectedTab) ? "true" : "false");
    propertyValue.push(tabList[index].ordinal);

    var resource = resourcePrefix + ":tab-" + (index + startIndex);
    // dump("\nresource: " + resource);
    var newNode = RDF.GetResource(resource);

    // Store general properties
    for (var i = 0; i < 8; i++) {
      if (propertyValue[i] != defaultValues[i]) { // Store modified values only!
        aResource = mzRDF + propertyArray[i];
        aDataSource.Assert(newNode, RDF.GetResource(aResource), RDF.GetLiteral(propertyValue[i]), true);
      }
    }
    // Store docShell properties
    for (i = 8; i < 12; i++) {
      propertyValue[0] = browser.docShell[propertyArray[i]].valueOf() ? "true" : "false";

      if (propertyValue[0] != defaultValues[i]) {  // Store modified values only!
        aResource = mzRDF + propertyArray[i];
        aDataSource.Assert(newNode, RDF.GetResource(aResource), RDF.GetLiteral(propertyValue[0]), true);
      }
    }
    tabContainer.AppendElement(newNode, true);
    // if (skipAboutBlank)
      // aDataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
    propertyValue = null;
  }
  return index;
}

function writeSessionName(aDataSource, aResource, aName)
{
  var property = RDF.GetResource("http://multizilla.mozdev.org/rdf#name");
  var currentValue = aDataSource.GetTarget(RDF.GetResource(aResource), property, true) || null;
  var newValue = RDF.GetLiteral(aName);

  addOrChangeValue(aDataSource, RDF.GetResource(aResource), property, currentValue, newValue);
  mzFlushDataSource(aDataSource);
}

function renameTabSession(aDataSource, aResource)
{
  var sessionName = aResource;
  var property = RDF.GetResource("http://multizilla.mozdev.org/rdf#name");
  var data = aDataSource.GetTarget(RDF.GetResource(aResource), property, true);

  if (data) {
    data = data.QueryInterface(Components.interfaces.nsIRDFLiteral);
    sessionName = data.Value;
  }
  window.openDialog("chrome://multiviews/content/tabsession-manager/tabSessionRenamingWindow.xul", "prompt",
                    "chrome, modal, dialog, centerscreen", window, aResource, sessionName, "");

  if (gSessionName) // && gSessionName != sessionName)
    writeSessionName(aDataSource, aResource, gSessionName);
}

function restoreTabSession()
{
  var windowResourceID = getRelativeWindowResourceID(window); // getWindowResourceID(window);

  if (windowResourceID == null) {
    alert("Error in restoreTabSession()");
    return;
  }
  var sessionID = 0;
  // Use a relative datasource ID to build the resource
  var resource = "urn:session-"+sessionID+":"+windowResourceID+":tabs:root";
  // window.document.documentElement.setAttribute("sessionID", resource); // Checked when saving tab sessions

  if (checkResource(mzRDFStore, resource)) {
    var windowManagerDS = Components.classes['@mozilla.org/rdf/datasource;1?name=window-mediator'].getService(nsIWindowDataSource);  
    // We used a relative datasource ID to build the resource but we need to use the 
    // original datasource ID of our window or 'getWindowResourceID' will return null!
    windowResourceID = getWindowResourceID(window);
    var browserWin = windowManagerDS.getWindowForResource(windowResourceID);
    var browser = browserWin.getBrowser();
    restoreTabsFromResource(mzRDFStore, browser, resource, false);
  }
  else {
    var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    promptService.alert(window, mzGetTextFromBundle("mzWizard"), mzGetTextFromBundle("noTabSessionAssociated"));
  }
}

function restoreBrowserSession(aCheckFirstWindow)
{
  if ((!isFirstWindow() && aCheckFirstWindow) || !mzRDFStore)
    return;
  // We can only check for duplicates *after* the startup page(s) is/are loaded!
  if (readMyPref("multizilla.prevent.double", "bool", true) || 
      readMyPref("browser.startup.page", "int", 0) > 0) {
    var delayed = false;

    if (pref.getIntPref("browser.startup.page") == 1) {
      var pageCount = pref.getIntPref("browser.startup.homepage.count");
      var tab = gBrowser.mTabContainer.childNodes;

      if (pageCount != tab.length)
        delayed = true;
      for (var i = 0; i < tab.length; i++) {
        if (tab[i].hasAttribute("busy")) {
          delayed = true;
          break;
        }
      }
    }
    if (delayed && aCheckFirstWindow) {
      setTimeout(restoreBrowserSession, 1000, true);
      return;
    }
  }
  var restoreAll = readMyPref("multizilla.tabs.restore-session.behavior", "int", 3);
  var skipSingleWindowConfirmation = readMyPref("multizilla.browsersession.skip.confirm-single-window", "bool", false);
  var windows = getWindowResourcesForSession(mzRDFStore, "session-0");
  var multipleWindows = (windows.length > 1);
  var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);

  if (restoreAll == 3 && (multipleWindows || (!skipSingleWindowConfirmation && windows.length == 1))) {
    var checkbox = {value:0};
    var text, windowButtonLabel, allWindowsButtonLabel;
    var title = mzGetTextFromBundle("restoreTabSessionTitle");
    // Are we restoring a single window?
    if (windows.length == 1) {
      // Yes, select text and button label for single window restore
      text = mzGetTextFromBundle("restoreTabSessionText");
      windowButtonLabel = mzGetTextFromBundle("restoreSessionButtonLabel");
      allWindowsButtonLabel = null; // Hide 'All Windows' button
    }
    else {
      // Select text and button label for multiple windows
      text = mzGetTextFromBundle("restoreAllSessionsText");
      text = text.replace(/%S/, windows.length);
      windowButtonLabel = mzGetTextFromBundle("restoreFirstWindowButtonLabel");
      allWindowsButtonLabel = mzGetTextFromBundle("restoreAllWindowsButtonLabel");
    }
    var result = promptService.confirmEx(window, title, text, 
                                         (promptService.BUTTON_POS_0 * promptService.BUTTON_TITLE_IS_STRING) +
                                         (promptService.BUTTON_POS_1 * promptService.BUTTON_TITLE_CANCEL) +
                                         (promptService.BUTTON_POS_2 * promptService.BUTTON_TITLE_IS_STRING),
                                         windowButtonLabel, null, allWindowsButtonLabel, null, checkbox);
    if (result == 1) // 0 = first window only / 1 = Cancel, Esc or close window / 2 = all windows
      return;
    else if (result == 0)
      restoreAll = 1;
  }
  // Restore all tabs from a priorly saved browser session
  var index = 0;
  var hasNext = true;
  var browser = getBrowser();
  var restoredTabs = 0;

  while (hasNext && index < windows.length) {
    if (index > 0 && restoreAll == 1) // Restore first window only!
      break;
    else {
      var resource = windows[index];

      if (checkResource(mzRDFStore, resource)) {
        if (index > 0)
          browser = openWindowAndReturnBrowser();
        // window.document.documentElement.setAttribute("sessionID", aResource);
        hasNext = restoreTabsFromResource(mzRDFStore, browser, resource, true);

        if (hasNext > 0) {
          restoredTabs += hasNext;
          index++;
        }
      }
      else 
        hasNext = false;
    }
  }
  var shouldNotify = readMyPref("multizilla.tabs.restore-session.notify", "bool", false)

  if (shouldNotify && windows.length >= 1) {
    var paramArray = [restoredTabs, index];
    var message = mzGetTextFromBundle("restoreTabSessionNotifyText", paramArray, null);
    message = message + " " + (index > 0 ? mzGetTextFromBundle("windowText") : mzGetTextFromBundle("windowsText"));
    var notifyText = mzGetTextFromBundle("mzWizard");
    var mostRecentWin = getMostRecentWindow();
    setTimeout(promptService.alert, 1000, mostRecentWin, notifyText, message);
  }
}

function restoreTabsFromResourceList(aDataSource, aBrowser, aResourceList)
{
  var resources = aResourceList.split(',');
  resources[0] = resources[0].replace('sm:resources:', '');

  for (i in resources) {
    var result;

    if (resources[i].match(':tabs:root')) { // Window
      result = restoreTabsFromResource(aDataSource, aBrowser, resources[i], true);
    }
    else { // Tab
	  var resource = RDF.GetResource(resources[i]);
      result = restoreTabFromResource(aDataSource, aBrowser, resource, true, false);
    }
    if (result == -1)
      break;
  }
  return result;
}

function restoreTabsFromResource(aDataSource, aBrowser, aResource, aStartupFlag)
{
  var tabContainer = mzRDFCUtils.MakeSeq(aDataSource, RDF.GetResource(aResource)).GetElements();

  if (!tabContainer)
    return -1;

  var addTab = true; // default is to add new tabs
  // The next two vars are used when we're called from restoreTabSession()
  var newTabs = true;
  var closeRemaining = false;

  if (aStartupFlag) { // called from restoreBrowserSession()
    // Do we need to use unused (blank) tabs before opening new tabs?
    if (useBlankTabs) {
      var useBlankTabs = readMyPref("multizilla.tabs.blank-tab-first", "bool", true);
      /*** 
        * The user wants to use unused (blank) tabs before opening new tabs. We *must* 
        * check the preference to find out if mozilla had to load one or more pages.
        * We can't simply use the first unused (blank) tab if that's the case!
        */	
      addTab = (readMyPref("browser.startup.page", "int", 0) > 0);
    }
  }
  else { // called from restoreTabSession() or session manager
    var prefValue = readMyPref("multizilla.newtabfor.groupfolders", "int", 0); // 0 = reuse / 1 = new / 2 = asks
    closeRemaining = readMyPref("multizilla.bookmark.close-remaining", "bool", false);
    // gBrowser = aBrowser;

    if (prefValue == 2) {
      window.openDialog("chrome://multiviews/content/multiviewsAskSettings.xul", "_blank", "chrome,modal,centerscreen", gBrowser, closeRemaining);
      if (gBrowser.mNewTabs == -1)
        return -1;
      newTabs = (gBrowser.mNewTabs == 0) ? false : true;
      closeRemaining = gBrowser.mRemoveRemaining;
    }
    else 
      newTabs = (prefValue == 1);
  }
  mzTabIndex = 0; // global var!
  var tabCount = 0;
  var addedTabs = 0;
  // addTab is now either true or false, based on the users pref settings
  while (tabContainer.hasMoreElements()) {

    try {
      var resource = tabContainer.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
      mzTabIndex++;

      if (aStartupFlag)
        addTab = (mzTabIndex > 1) ? true : addTab
      else if (newTabs == false) {
        tabCount = gBrowser.mPanelContainer.childNodes.length;
        addTab = (mzTabIndex > tabCount);

        if (addTab)
          addedTabs++;
      }
      restoreTabFromResource(aDataSource, aBrowser, resource, addTab, aStartupFlag);
    } catch(ex) {
      mzDebug(ex);
      continue;
    }
  }
  // Do we need to remove any remaining tabs? 
  // Skip this section for invalid resources (mzTabIndex == 0)
  if (closeRemaining && mzTabIndex > 0) {
    tabCount = gBrowser.mTabContainer.childNodes.length;
    addedTabs = tabCount - mzTabIndex;

    // Do we have any new tabs?
    if (addedTabs == 0)
      return mzTabIndex; // Nope, nothing to do
    if (aStartupFlag) { // restoreBrowserSession (at startup) only!
      // var startTab = newTabs ? (tabCount+1) : 0;
      var i = 0;

      if (newTabs) {
        /* startTab = gBrowser.mTabList[startTab].ordinal-1;
        for (i = startTab; i >= 0; i--)
          gBrowser.removeTab(gBrowser.mTabList[i]); */
      }
      else {
        for (i = tabCount; i >= mzTabIndex; i--)
          gBrowser.removeTab(gBrowser.mTabList[i]);
      }    
    }
  }
  return mzTabIndex;
}

function addOrChangeValue(aDataSource, aResource, aProperty, currentValue, aValue)
{
  if (currentValue)
    aDataSource.Change(aResource, aProperty, currentValue, aValue);
  else
    aDataSource.Assert(aResource, aProperty, aValue, true);
  mzFlushDataSource(aDataSource);
}

function restoreTabFromResource(aDataSource, aBrowser, aResource, aNewTabFlag, aStartupFlag)
{
  var dataCollection = getTabDataFromResource(aDataSource, aResource);

  if (aBrowser) {
    addTabFromCollection(aBrowser, dataCollection, aNewTabFlag, aStartupFlag);
  }
}

function getTabDataFromResource(aDataSource, aResource)
{
  var dataCollection = new Array;
  var propertyArray = ["label", "custom", "firstpage", "image", "location", "interval", "selected", "ordinal", "allowPlugins", "allowJavascript", "allowMetaRedirects", "allowSubframes", "allowImages"];
  var defaultValues = ["about:blank", "false", "about:blank", "", "about:blank", "0", "false", "0", "true", "true", "true", "true", "true"];
  // const addMissingProperties = false;
  
  for (var i = 0; i < propertyArray.length; i++) {
    if (i == 7) // Skip ordinal!
      continue;
    var resource = "http://multizilla.mozdev.org/rdf#"+propertyArray[i];
    var property = RDF.GetResource(resource);
    var collection = new Object();

    try {
      var data = aDataSource.GetTarget(aResource, property, true);
      data = data.QueryInterface(Components.interfaces.nsIRDFLiteral);
      collection.value = data.Value;
      dataCollection[i] = collection;
    }
    catch(ex) {
      collection.value = defaultValues[i];
      dataCollection[i] = collection;

      /* if (addMissingProperties) {
        alert("Error: "+propertyArray[i]+" is missing...adding with default value: "+collection.value);
        aDataSource.Assert(aResource, property, RDF.GetLiteral(collection.value), true);
        aDataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
      } */
    }
  }
  return dataCollection;
}
//////////////////////////////////////////////////////////////////////////
// This function opens a new tab, or uses a blank tab, for the given URL 
// and sets all attributes and docShell properties from 'aDataCollection'
//////////////////////////////////////////////////////////////////////////
function addTabFromCollection(aBrowser, aDataCollection, newTabFlag, aStartupFlag)
{
  var URL = aDataCollection[4].value;
  var preventDuplicatedTabs = readMyPref("multizilla.prevent.double", "bool", true);

  if (URL == "about:blank" || (preventDuplicatedTabs && preventDupeTab(URL, aStartupFlag, aBrowser)))
    return;

  var browser, tab = aBrowser.selectedTab;

  if (newTabFlag) {
    browser = checkForBlankTab(aBrowser);

    if (!browser) {
      // 'restoreTab' skips setting the default permissions in addOrInsertTab
      tab = aBrowser.addTab(null, null, null, null, "restoreTab");
    }
  }
  else if (aStartupFlag == false) {
    tab = aBrowser.mTabList[mzTabIndex-1];
  }
  if (!browser) 
    browser = aBrowser.getBrowserForTab(tab);
  // Set browser properties before we load the URL for this tab
  browser.docShell.allowPlugins = (aDataCollection[8].value == "true") ? true : false;
  browser.docShell.allowJavascript = (aDataCollection[9].value == "true") ? true : false;
  browser.docShell.allowMetaRedirects = (aDataCollection[10].value == "true") ? true : false;
  browser.docShell.allowSubframes = (aDataCollection[11].value == "true") ? true : false;
  browser.docShell.allowImages = (aDataCollection[12].value == "true") ? true : false;

  // Do we have a custom label?
  if (aDataCollection[1].value == "true") { // Yes, restore custom label
    tab.setAttribute("customLabel", true);
    tab.setAttribute("label", aDataCollection[0].value);
  }
  /***
    * Set 'firstPage' attribute (used for menu item Tab Home and keyboard shortcut Shift+Home)
    * Note that function preventDupeTab() might fail when 'firstPage' is set to aDataCollection[2].value
    * (special case only) so that is why we have to check 'preventDuplicatedTabs'
    */
  tab.setAttribute("firstPage", preventDuplicatedTabs ? aDataCollection[4].value : aDataCollection[2].value);
  // The properties are set so lets load the URL
  browser.loadURI(URL);

  var interval = aDataCollection[5].value;

  if (interval >= 30000) // We only honor intervals of 30+ seconds
    aBrowser.setReloadInterval(tab, interval);
  // Activate the (new) tab
  if (aDataCollection[6].value == "true")
    aBrowser.selectedTab = tab;
}

function removeAllResourcesFor(aDataSource, aResource)
{
  removeResources(aDataSource, aResource);
  removeDataResources(aDataSource, aResource);
  removeSession(aDataSource, aResource);
  mzFlushDataSource(aDataSource);
}

function removeTabResourcesFor(aDataSource, aResource) 
{
  removeResources(aDataSource, aResource);
  removeDataResources(aDataSource, aResource);

  var tabContainerResource = aResource.replace(/:tabs:tab-\w*/, ":tabs:root");
  var tabContainer = mzConstructRDFContainer(aDataSource, "seq", tabContainerResource, false);

  tabContainer.RemoveElement(RDF.GetResource(aResource), true);
  removeSession(aDataSource, aResource);
  mzFlushDataSource(aDataSource);
}

function removeResources(aDataSource, aResource)
{
  var tabContainer = mzRDFCUtils.MakeSeq(aDataSource, RDF.GetResource(aResource)).GetElements();

  if (!tabContainer)
    return;
  mzDebug("removeResources");
  /*** 
  * EXAMPLE: removeResources('urn:session-0:window-0:tabs:tab-0');
  * removes:
    <RDF:Description about="urn:session-0:window-0:tabs:tab-0"
                     MZ:label="mozdev.org - multizilla: index"
                     MZ:custom="false"
                     MZ:firstpage="false"
                     MZ:image="http://www.mozilla.org/images/mozilla-16.png"
                     MZ:location="http://multizilla.mozdev.org/index.html"
                     MZ:interval="0"
                     MZ:selected="false"
                     MZ:allowPlugins="true"
                     MZ:allowJavascript="true"
                     MZ:allowMetaRedirects="true"
                     MZ:allowSubframes="true"
                     MZ:allowImages="true" /> */

  while (tabContainer.hasMoreElements()) {
    try {
      var resource = tabContainer.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
      var labels = aDataSource.ArcLabelsOut(resource);

      while (labels.hasMoreElements()) {
        try {
          var label = labels.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
          var data = aDataSource.GetTarget(resource, label, true);
          aDataSource.Unassert(resource, label, data);
        } 
        catch(ex) { 
          mzDebug(ex);
        }
	  }
    } catch(ex) {
      mzDebug(ex);
      continue;
    }
  }
}

function removeDataResources(aDataSource, aResource)
{
  var resource = RDF.GetResource(aResource);
  var labels = aDataSource.ArcLabelsOut(resource);
  /*** 
  * EXAMPLE: removeDataResources('urn:session-0:window-0:tabs:root');
  * removes:
    <RDF:Seq about="urn:session-0:window-0:tabs:root">
      <RDF:li resource="urn:session-0:window-0:tabs:tab-0"/>
    </RDF:Seq> */

  while (labels.hasMoreElements()) {
    try {
      var label = labels.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
      var data = aDataSource.GetTarget(resource, label, true);
      aDataSource.Unassert(resource, label, data);
    }
    catch(ex) {
      mzDebug(ex);
    }
  }
}

function removeSession(aDataSource, aResource)
{
  var sessionsContainer = getSessionContainer(aDataSource);
  sessionsContainer.RemoveElement(RDF.GetResource(aResource), true);
  /***
    * We need to cleanup multizilla.rdf when we remove the last window:
    *   <RDF:Seq about="urn:sessions:root">
    *   </RDF:Seq>
    * ...or we keep the above 'garbage' alive in multizilla.rdf
    */
  if (sessionsContainer.GetCount() < 1) {
    removeDataResources(aDataSource, "urn:sessions:root");
  }
}

function checkResource(aDataSource, aResource)
{
  var tabContainer = mzRDFCUtils.MakeSeq(aDataSource, RDF.GetResource(aResource)).GetElements();
  var found = tabContainer.hasMoreElements() ? true : false;
  if (!found)
    removeDataResources(aDataSource, aResource); // cleanup junk resource
  return found;
}
////////////////////////////////////////////////////////////////////////////////
// example: 
// checkResourceForValue("urn:sessions:root", "urn:session-0:window-1:tabs:root")
////////////////////////////////////////////////////////////////////////////////
function checkResourceForValue(aDataSource, aResourceID, aValue)
{
  var container = mzRDFCUtils.MakeSeq(aDataSource, RDF.GetResource(aResourceID)).GetElements();

  if (!aDataSource || !container)
    return false;

  while (container.hasMoreElements()) {
    try {
      var resource = container.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
      if (resource.Value == aValue)
        return resource.Value;
    } catch(ex) {
      mzDebug(ex);
    }
  }
  return false;
}

function mzGetLastResourceFromContainer(aDataSource, aResource)
{
  if (aResource == undefined)
    throw("aResource == undefined in mzGetLastResourceFromContainer()");
  else if (typeof aResource == 'string')
    aResource = RDF.GetResource(aResource);

  var resource = null;
  var container = mzRDFCUtils.MakeSeq(aDataSource, aResource).GetElements();

  if (!container)
    return -1;

  while (container.hasMoreElements()) {
    resource = container.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
  }
  return (resource == null) ? null : resource.Value;
}

function mzGetParentOfContainer(aDataSource, aContainer)
{
  aContainer = RDF.GetResource(aContainer);
  var arcsIn = aDataSource.ArcLabelsIn(aContainer);

  while (arcsIn.hasMoreElements()) {
    var containerArc = arcsIn.getNext();
    if (mzRDFCUtils.IsOrdinalProperty(containerArc)) {
      /***
        * "urn:session-x:window-x:tabs:tab-x" -> "urn:session-x:window-x:tabs:root"
        * "urn:session-x:window-x:tabs:root" -> "urn:sessions:root"
        */
      var resource = aDataSource.GetSources(containerArc, aContainer, true).getNext().QueryInterface(Components.interfaces.nsIRDFResource);
      // alert(resource.Value);
    }
  }
  return resource.Value || null;
}
////////////////////////////////////////////////////////////////////////////////
// This function reads from a given file (aFile) and returns a pointer the 
// content or null on failure
////////////////////////////////////////////////////////////////////////////////	
function mzReadFromFile(aFile) 
{
  if (aFile) {
    var inputStream = Components.classes["@mozilla.org/network/file-input-stream;1"]
                                .createInstance(Components.interfaces.nsIFileInputStream);

    var scriptableStream = Components.classes["@mozilla.org/scriptableinputstream;1"]
                                     .createInstance(Components.interfaces.nsIScriptableInputStream);

    try {
      /**
       * inputStream.init( file, ioFlags, perm, behaviorFlags );
       *
       * @param file          - file to read from (must QI to nsILocalFile)
       * @param ioFlags       - file open flags (see File open bits)
       * @param permission    - file mode bits (see File mode bits)
       * @param behaviorFlags - flags specifying various behaviors of the class (see enumerations in the class)
      */
      inputStream.init(aFile, 1, 0, 0);
      scriptableStream.init(inputStream);
      var content = scriptableStream.read(scriptableStream.available());
      scriptableStream.close();
      inputStream.close();
    } catch(ex) {
      mzDebug(ex);
      return null;
    }
    return content;
  }
  return null;
}
////////////////////////////////////////////////////////////////////////////////
// This function write the given content (aContent) to a given file (aFile) and 
// it returns false on failure and true on success.
////////////////////////////////////////////////////////////////////////////////
function mzWriteToFile(aFile, aContent)
{
  if (aFile && aContent) {
    if (aFile.exists()) 
      aFile.remove(true);

    var outputStream = Components.classes["@mozilla.org/network/file-output-stream;1"]
                                 .createInstance(Components.interfaces.nsIFileOutputStream);

    try {
      aFile.create(aFile.NORMAL_FILE_TYPE, 0666);
      /**
       * outputStream.init( file, ioFlags, perm, behaviorFlags );
       *
       * @param file          - file to write to (must QI to nsILocalFile)
       * @param ioFlags       - file open flags (see File open bits)
       * @param permission    - file mode bits (see File mode bits)
       * @param behaviorFlags - currently none supported
      */
      outputStream.init(aFile, 2, 0x200, 0);
      outputStream.write(aContent, aContent.length);
      outputStream.close();
    } catch(ex) {
      mzDebug(ex);
      return false;
    }
    return true;
  }
  return false;
}
////////////////////////////////////////////////////////////////////////
function mzFlushDataSource(aDataSource)
{
  try {
    aDataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
  } catch(ex) {
    // die silently
  }
}
////////////////////////////////////////////////////////////////////////
// Debug helper
////////////////////////////////////////////////////////////////////////
function mzDebug(txt)
{
  if (gDebug)
    dump("MultiZilla: " + txt + "\n");
}

function testCode(aBookmarkID)
{
  return;
}